/*
 .
 .		Name:				tzlib.c
 .
 .		Description:		Time/Date conversion library.
 .
 .		Version:			1.0 - Initial version
 .
 .							1.1 - Corrections to UTCToLocalTime() for correct
 .								determination of daylight time vs. standard
 .								time.
 .								- Added IsDSTDefined() routine
 .
 .							1.2 - Fixed UTCToLocalTime() to properly deal with
 .								time zone descriptions that did not define DST.
 .
 .							1.3 - Added sliding window year conversion
 .								- Added day of week and month descriptions
 .
 .							1.4 - Fixed GetSecondsSince1980() so that it works properly
 .                                when date in WCATIME is not defined
 .
 .							1.5 - Added GetAmPmHour() function
 .
 .							1.6 - Added Win32ToWCATime() function
 .
 .							1.7 - Added a leap-year days of month elapsed table
 .
 .							1.8 - Added GetSecondsSince1900(), GetDaysSince1900()
 .                                functions.
 .
 .                          1.9 - UTCToLocalTime() modified to support time zones for
 .                                which daylight savings time crosses over to the
 .                                following year.
 .                              - Removed the IsDST() function because by original
 .                                definition it could not precisely determine
 .                                daylight/standard time one hour before the event
 .                                is to occur.
 .                              - Casted some assignments to eliminate compiler warnings.
 .							2.0 - Added some conditional compilation directives and code
 .								  comments to allow use with Dynamic C compiler.
 .							    - Adds Win32ToWCATimeZoneInfo() call.
 .                 3/9/2012 2.1 - Corrected Leap Year bug in GetDaysSince1900() and GetDaysSince1980()
 .
 .
 */

//#ifndef TRUE
//#define TRUE   1
//#define FALSE  0
//#endif


#include "StdAfx.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys\types.h>

#include "tzlib.h"

#include "DebugOutput.h"

// Instead, under properties of this file, turn off /clr
//#pragma unmanaged

const unsigned short _daysMonthEachNonLeapYear[13] = {
	0,
   31,   /* January   */
   28,   /* February  */
   31,   /* March     */
   30,   /* April     */
   31,   /* May       */
   30,   /* June      */
   31,   /* July      */
   31,   /* August    */
   30,   /* September */
   31,   /* October   */
   30,   /* November  */
   31    /* December  */
};

const unsigned short _daysMonthEachLeapYear[13] = {
   0,
   31,   /* January   */
   29,   /* February  */
   31,   /* March     */
   30,   /* April     */
   31,   /* May       */
   30,   /* June      */
   31,   /* July      */
   31,   /* August    */
   30,   /* September */
   31,   /* October   */
   30,   /* November  */
   31    /* December  */
};

const unsigned short _daysMonthElapsedNonLeapYear[13] = {
	0,
	31,		/* through January									*/
	59,		/* through February (not including any leap year)	*/
	90,		/* through March									*/
	120,	/* through April									*/
	151,	/* through May										*/
	181,	/* through June										*/
	212,	/* through July										*/
	243,	/* through August									*/
	273,	/* through September								*/
	304,	/* through October									*/
	334,	/* through November									*/
	365		/* through December									*/
};

const unsigned short _daysMonthElapsedLeapYear[13] = {
	0,
	31,		/* through January									*/
	60,		/* through February (this and rest include leap		*/
	91,		/* through March									*/
	121,	/* through April									*/
	152,	/* through May										*/
	182,	/* through June										*/
	213,	/* through July										*/
	244,	/* through August									*/
	274,	/* through September								*/
	305,	/* through October									*/
	335,	/* through November									*/
	366		/* through December									*/
};

const unsigned char _monthsDesc1[12][4] = {
	"Jan","Feb","Mar","Apr","May","Jun",
	"Jul","Aug","Sep","Oct","Nov","Dec"
};

const unsigned char _monthsDesc2[12][12] = {
	"January","February","March",
	"April",  "May",     "June",
	"July",   "August",  "September",
	"October","November","December"
};

const unsigned char _daysDesc1[7][4] = {
	"Sun","Mon","Tue","Wed","Thu","Fri","Sat"
};

const unsigned char _daysDesc2[7][12] = {
	"Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"
};

const unsigned char _ampmDesc1[2][3] = {
	"AM", "PM" 
};

const unsigned char _ampmDesc2[2][3] = {
	"am", "pm"
};



/*
 .    Name:          IsDSTDefined()
 .
 .    Description:   Tests if daylight savings time has been defined for a
 .                   given time zone description.
 .
 .    In:            'tz' - time zone to test
 .
 .    Returns:       TRUE if daylight savings time defined.
 .
 */

unsigned char IsDSTDefined(WCA_TIME_ZONE_INFORMATION *tz)
{
   if ((!tz->StandardDate.wMonth) || (!tz->DaylightDate.wMonth))
      return(FALSE);
   else
      return(TRUE);
}


/*
 .    Name:          LocalTimeToUTC()
 .
 .    Description:   Converts local time/date as defined by passed time zone
 .                   information structure to UTC time.
 .
 .    In:            'local' - local time/date to convert
 .                   'tz'    - time zone information defining local time
 .                   'utc'   - the local time/date 'l' converted to UTC 
 .
 .		Note:				This routine *REQUIRES* the 'wIsDstActive' member of WCATIME to be set
 .							to correctly interpret the local time when converting to UTC.
 */

unsigned short LocalTimeToUTC(WCATIME *myLocal, WCA_TIME_ZONE_INFORMATION *tz, WCATIME *myUtc)
{
	unsigned long secs;

	secs = GetSecondsSince1980(myLocal);

	// keep the math using unsigned longs
	if (tz->Bias >= 0L)
	{
		// positive bias, add to get UTC
		secs = secs + (unsigned long)(tz->Bias * 60L);
	} else
	{
		// negative bias, subtract to get UTC
		secs = secs - (unsigned long)(tz->Bias * (-60L));
	}

	if (myLocal->wIsDstActive && IsDSTDefined(tz))
	{

		/* local time is flagged as in daylight savings time, we remove any DST bias */
		// keep the math using unsigned longs
		if (tz->DaylightBias >= 0L)
		{
			// positive bias, add to get UTC
			secs = secs + (unsigned long)(tz->DaylightBias * 60L);
		} else
		{
			// negative bias, subtract to get UTC
			secs = secs - (unsigned long)(tz->DaylightBias * (-60L));
		}
	}
	else
	{
		/* local time is flagged as in standard time, we remove any standard bias */
		/* local time is flagged as in daylight savings time, we remove any DST bias */
		// keep the math using unsigned longs
		if (tz->StandardBias >= 0L)
		{
			// positive bias, add to get UTC
			secs = secs + (unsigned long)(tz->StandardBias * 60L);
		} else
		{
			// negative bias, subtract to get UTC
			secs = secs - (unsigned long)(tz->StandardBias * (-60L));
		}
	}

	GetTimeDateFromSecondsSince1980(secs,myUtc);

	/* set 'wIsDstActive' member FALSE since UTC time has no daylight savings time */

	myUtc->wIsDstActive = 0;

   /* pass milliseconds */

	myUtc->wMilliseconds = myLocal->wMilliseconds;


	return(TRUE);
}



/*
 .    Name:          UTCToLocalTime()
 .
 .    Description:   Converts UTC time to local time as defined by passed
 .                   time zone definition.  If passed time zone definition
 .                   declares daylight savings time then local time will be
 .                   adjusted accordingly.
 .
 .    In:            'utc'   - UTC time to convert
 .                   'local' - the UTC time 'utc' converted to local time
 .                   'tz'    - time zone information defining local time
 .
 .    Returns:       TRUE on success with local time in 'local'
 .                   FALSE on failure with contents of 'local' undefined
 */

unsigned short UTCToLocalTime(WCATIME *utc, WCA_TIME_ZONE_INFORMATION *tz, WCATIME *local)
{	
	unsigned long	dstStartUtc, dstEndUtc, utcSecs, secs;
	int dstActive;

	if ((utc->wHour > 23) || (utc->wMinute > 59) || (utc->wSecond > 59))
		return(FALSE);

	if ((utc->wMonth < 1) || (utc->wMonth > 12) || (utc->wDay < 1) || (utc->wDay > 31) || (utc->wYear < 1980))
		return(FALSE);

	local->wMilliseconds = utc->wMilliseconds;
	
	utcSecs = GetSecondsSince1980(utc);

	// keep the math using unsigned longs
	if (tz->Bias >= 0)
	{
		secs = utcSecs - (unsigned long)(tz->Bias * 60L);
	} else
	{
		secs = utcSecs + (unsigned long)(tz->Bias * (-60L));
	}
		
    if (IsDSTDefined(tz))
    {
		// The dstStart and dstEnd are referenced to UTC time.
		// ie. the dstStart and dstEnd are not based on local time, they are based on UTC time
		// So, if DST starts at 2:00 am local time with a time offset of -6:00:00
		// (US central time), the time of the dst Start would be 8:00 am UTC.

		dstStartUtc = BuildDSTSwitchTime(TRUE, tz, utc);

		dstEndUtc   = BuildDSTSwitchTime(FALSE, tz, utc);

		if ((utcSecs >= dstStartUtc) && (utcSecs < dstEndUtc))
		{
			// we are in daylight savings time
			// add the daylight savings offset
		    if (tz->DaylightBias >= 0)
			{
				secs = secs + (unsigned long)(tz->DaylightBias * (-60L));
			} else
			{
				secs = secs - (unsigned long)(tz->DaylightBias * 60L);
			}

			dstActive = 1;
		}
		else
		{
			dstActive = 0;
		}
	}
	else
	{
		dstActive = 0;
	}

	local->wIsDstActive = dstActive;


    GetTimeDateFromSecondsSince1980(secs,local);

	return(TRUE);
}


/* 
 .		Name:			GetSecondsSince1980()
 .
 .		Description:	Get seconds since January 1, 1980 to current time/date
 .
 */

unsigned long GetSecondsSince1980(WCATIME *time)
{
	unsigned long retval;

	if (!time->wMonth || !time->wDay || !time->wYear)
	{
		// date information isn't provided (or not provided correctly, so skip the
		// call to GetDaysSince1980() effectively returning seconds elapsed since 00:00:00
		// of current day

		retval = 0L;
	}
	else
	{
		retval  = GetDaysSince1980(time,TRUE) - 1UL;	/* don't count the current day */
		if (retval == 0xFFFFFFFFUL)
			return retval;

		retval *= 86400UL;							    /* convert to seconds */
	}
	
	retval += (((unsigned long)time->wHour * 3600UL) + ((unsigned long)time->wMinute * 60UL) + ((unsigned long)time->wSecond));	/* add the current time of day */

	return(retval);
}


/* 
 .		Name:			GetDaysSince1980()
 .
 .		Description:	Returns number of days elapsed since January 1, 1980.  If
 .						'thisMonth' is FALSE GetDaysSince1980() will not add
 .						the days elapsed in the current month to the value
 .						it returns.
 .
 .		Note:			The current day (although not necessarily elapsed) is also included
 .						in the value returned.
 .
*/

unsigned long GetDaysSince1980(WCATIME *timeIn, unsigned short includeThisMonth)
{
	unsigned long ulRetval;
	unsigned long ulDays;
	unsigned long ulYear;

	WCATIME *lTime;

	lTime = timeIn;

	if ((lTime->wMonth < 1) || (lTime->wMonth > 12))
		return(0xFFFFFFFFUL);

	/* account for days in years since 1980
	   note: this simple formula works because year 2000 happens to be a leap year
	*/

	ulYear = (UINT32)lTime->wYear - 1980ul;
	ulDays = (ulYear * 365ul) + (ulYear>>2) + 1ul;

	ulRetval = ulDays;

	/* account for days in months this year */

	ulRetval += _daysMonthElapsedNonLeapYear[lTime->wMonth-1];

		
	/* account for days in months this year */

	if (IsLeapYear(lTime->wYear) && (lTime->wMonth < 3))
	{
		ulRetval--;
	}

	/* account for days in the current month */

	if (includeThisMonth)
		ulRetval += (unsigned long)lTime->wDay;

	return ulRetval;
}




/*
 .		Name:				GetTimeDateFromSecondsSince1980()
 .
 .		Description:		Convert seconds since January 1, 1980 to current time/date
 .							in WCATIME structure.
 .
 .		Note:				This routine *CANNOT* determine if converted time is within
 .							daylight savings time!
 */

void GetTimeDateFromSecondsSince1980(unsigned long secs, WCATIME *dateOut)
{
	unsigned long leapDays1, leapDays2;
	unsigned long datePart;
	unsigned long timePart;
	unsigned short	daysLeft;
	unsigned short	tt, xx;

	datePart = secs / 86400UL;
	timePart = secs % 86400UL;

	dateOut->wDayOfWeek = GetDayOfWeek(datePart + 1);

	switch(datePart)
	{
		case 0:
			dateOut->wMonth = 1;
			dateOut->wDay   = 1;
			dateOut->wYear  = 1980;
			break;

		default:
			leapDays1 = leapDays2 = 1UL;

			if (datePart > 366UL)
			{
				leapDays1 += (datePart - 365UL) / 1461UL; // (4*365)+1 == 1461
				leapDays2 += datePart / 1461UL;
			}
						
			dateOut->wYear = (unsigned short)(1980UL + ((datePart - leapDays1) / 365UL));

			if (dateOut->wYear == 1980)
				daysLeft = (unsigned short)(datePart + 1UL);
			else
				daysLeft = (unsigned short)(1UL + (unsigned long)IsLeapYear(dateOut->wYear) + datePart - leapDays2 - (365UL * ((unsigned long)dateOut->wYear - 1980UL)));
				
			for(tt = 1; tt <= 12; tt++)
			{
				xx = GetDaysInMonth(tt,dateOut->wYear);

				if (daysLeft > xx) 
				{
					daysLeft -= xx;
				}
				else
				{
					dateOut->wMonth = tt;
					dateOut->wDay = daysLeft;
					break;
				}
			}
	}		
	
	dateOut->wHour   = (unsigned short)(timePart / 3600UL);
	dateOut->wMinute = (unsigned short)((timePart % 3600UL) / 60UL);
	dateOut->wSecond = (unsigned short)((timePart % 3600UL) % 60UL);

	/* because we don't know if the converted time is in DST or not */

	dateOut->wIsDstActive = 0;
}


/* 
 .		Name:			GetSecondsSince1900()
 .
 .		Description:	Get seconds since 00:00:00 January 1, 1900 to current time/date
 .
 */

unsigned long GetSecondsSince1900(WCATIME *time)
{
	unsigned long retval;

	if (!time->wMonth || !time->wDay || !time->wYear)
	{
		// date information isn't provided (or not provided correctly, so skip the
		// call to GetDaysSince1900() effectively returning seconds elapsed since 00:00:00
		// of current day

		retval = 0UL;
	}
	else
	{
		retval  = GetDaysSince1900(time,TRUE) - 1UL;		/* don't count the current day */
		retval *= 86400UL;							/* convert to seconds */
	}
	
	retval += (((unsigned long)time->wHour * 3600UL) + ((unsigned long)time->wMinute * 60UL) + (unsigned long)(time->wSecond));	/* add the current time of day */

	return(retval);
}


/* 
 .		Name:			GetDaysSince1900()
 .
 .		Description:	Returns number of days elapsed since January 1, 1900.  If
 .						'thisMonth' is FALSE GetDaysSince1900() will not add
 .						the days elapsed in the current month to the value
 .						it returns.
 .
 .		Note:			The current day (although not necessarily elapsed) is also included
 .						in the value returned.
 .
*/

unsigned long GetDaysSince1900(WCATIME *time, unsigned short thisMonth)
{
	unsigned long ulRetval;
	unsigned long ulDays;
	unsigned long ulYear;

	if ((time->wMonth < 1) || (time->wMonth > 12))
		return 0xffffffffUL;

   /* account for days in years since 1900
	   note: this simple formula works because year 2000 happens to be a leap year
	 */

	ulYear   = (UINT32)time->wYear - 1900ul;
	ulDays   = (ulYear * 365ul) + (ulYear>>2);
	ulRetval = ulDays;

		
	/* account for days in months this year */

	ulRetval += _daysMonthElapsedNonLeapYear[time->wMonth-1];

	if (IsLeapYear(time->wYear) && (time->wMonth < 3))
		ulRetval--;

	/* account for days in the current month */

	if (thisMonth)
		ulRetval += (unsigned long)time->wDay;
	
	return(ulRetval);
}


/*
 .    Name:          IsDST()
 .
 .    Description:   Determines if a local time is within daylight savings
 .                   time based on time zone information structure.
 .
 .    Returns:       TRUE if time/date is within daylight savings time
 .                   FALSE if time/date is not within daylight savings time,
 .                   or if daylight savings time not defined for time zone.
 .
 .	  Note:			 this routine does not inspect or alter the contents of
 .					 the 'wIsDstActive' member of WCATIME.
 .
 */

// needs to be changed to receive local time not UTC time

//unsigned short IsDST(WCATIME *currentUtcTime, WCA_TIME_ZONE_INFORMATION *tz)
//{
//   unsigned long	dstStart, dstEnd, secs;
//   WCATIME			tmpTime;
//   
//   if (!(tz->DaylightDate.wMonth && tz->StandardDate.wMonth))
//   {
//      /* DST is not defined for this time zone */
//
//      return(0UL);
//   }
//
//   dstStart = BuildDSTSwitchTime(TRUE, &tz->DaylightDate,currentUtcTime);
//
//   dstEnd = BuildDSTSwitchTime(FALSE, &tz->StandardDate,currentUtcTime);
//
//   secs = GetSecondsSince1980(currentUtcTime);
//
//   if ((time >= dstStart) &&  (time < dstEnd))
//   {
//		return 1;
//	 } else
//	 {
//		 return 0;
//	 }
//}



// builds a DST switch time referenced to UTC time
// UTC time does not change for DST.
// returns 0 if DST not defined for time zone, otherwise returns the DST switch time in UTC seconds since 1980
//
// Note: at this time we can only do a DST shift that is reference to local time. If the standard wants to shift 
//       based on UTC time then the values used in the switchTime must be set accordingly.
//
unsigned long BuildDSTSwitchTime(bool bIsDaylighSwitch, WCA_TIME_ZONE_INFORMATION *tz, WCATIME *currentUtcTime)
{
	WCATIME			tmpTime;
	WCATIME         localSwitchTime;
	WCATIME        *switchTime;
	unsigned short	dayCnt,dayHitCnt, daysInMonth, xx;
	unsigned short  dstStartYear, dstEndYear, year;
	unsigned char	dayWeek;
	unsigned long   utcSwitchTime;

	if ((!tz->DaylightDate.wMonth) || (!tz->StandardDate.wMonth))
	{
		/* DST has not been defined for the specified time zone */
		return 0UL;
	}

	// If the daylight month is < standard month then the year stays the same.
	// else the year changes.
	if (tz->DaylightDate.wMonth < tz->StandardDate.wMonth)
	{
		//the year stays the same
		dstStartYear = dstEndYear = currentUtcTime->wYear;
	} else
	{
		// the dst start year is different from the dst end year.
		// We are in the southern hemisphere.
		// DST starts in the Oct-Nov and ends in Mar-May.

		// Have we in the first half of the year?
		if (currentUtcTime->wMonth < 6)
		{
			// yes
			// Has DST ended? If so, we calculate for the next year
			if (currentUtcTime->wMonth > tz->StandardDate.wMonth)
			{
				dstStartYear = currentUtcTime->wYear;
				dstEndYear   = currentUtcTime->wYear + 1;
			} else
			{
				dstStartYear = currentUtcTime->wYear - 1;
				dstEndYear   = currentUtcTime->wYear;
			}
		} else
		{
			// no
			dstStartYear = currentUtcTime->wYear;
			dstEndYear   = currentUtcTime->wYear + 1;
		}
	}

	if (bIsDaylighSwitch)
	{
		switchTime = &tz->DaylightDate;
		year       = dstStartYear;
	} else
	{
		switchTime = &tz->StandardDate;
		year       = dstEndYear;
	}

	localSwitchTime.wMonth        = switchTime->wMonth;		
	localSwitchTime.wYear         = year;
	localSwitchTime.wHour         = switchTime->wHour;		
	localSwitchTime.wMinute       = switchTime->wMinute;		
	localSwitchTime.wSecond       = switchTime->wSecond;		
	localSwitchTime.wMilliseconds = switchTime->wMilliseconds;

	// the only thing left is the date

	if (switchTime->wYear)
	{
		/* an absolute DST switch was specified, so we just fill in the the day of the month */
		localSwitchTime.wDay   = switchTime->wDay;		
	} else
	{
		tmpTime.wIsDstActive  = FALSE;
		tmpTime.wMilliseconds = 0;
		tmpTime.wHour         = 6;
		tmpTime.wMinute       = 0;
		tmpTime.wSecond       = 0;
		tmpTime.wDayOfWeek    = 0;
		tmpTime.wDay          = 1;
		tmpTime.wMonth        = switchTime->wMonth;
		tmpTime.wYear         = currentUtcTime->wYear;

		daysInMonth = GetDaysInMonth(switchTime->wMonth,currentUtcTime->wYear);

		// days to the first of the month the switch occurs in
		dayCnt = (unsigned short)GetDaysSince1980(&tmpTime,FALSE);

		if (switchTime->wDay == 5)
		{
			/* DST switches on last day of week Y        */
			/* for example, the last Sunday of the month */

			for(xx = daysInMonth; xx > 0; xx--)
			{
				dayWeek = GetDayOfWeek(dayCnt + xx);

				if (dayWeek == switchTime->wDayOfWeek)
				{
					localSwitchTime.wDay = xx;
					break;
				}
					
			}
		}
		else
		{
			if ((switchTime->wDay < 1) || (switchTime->wDay > 4))
				return(FALSE);

			/* DST switches on the first, second, third, or fourth day of week Y */
			/* For example, the second Sunday of the month                       */

			dayHitCnt = 0;
         
			for(xx = 1; xx <= daysInMonth; xx++)
			{
				dayWeek = GetDayOfWeek(dayCnt + xx);

				if (dayWeek == switchTime->wDayOfWeek)
				{
					dayHitCnt++;

					if (dayHitCnt == switchTime->wDay)
					{
						localSwitchTime.wDay = xx;
						break;
					}
				}
			}
		}
	}

	// We now have a date and time in local time the DST switch occurs on
	// This needs to be converted to UTC time

	utcSwitchTime = GetSecondsSince1980(&localSwitchTime);

	// adjust for the time zone bias
	if (tz->Bias >= 0L)
	{
		// add to get UTC
		utcSwitchTime = utcSwitchTime + (unsigned long)(tz->Bias * 60L);
	} else
	{
		// subtract to get UTC
		utcSwitchTime = utcSwitchTime - (unsigned long)(tz->Bias * (-60L));
	}

	// adjust for dst being active
	if (!bIsDaylighSwitch)
	{
		if (tz->DaylightBias >= 0L)
		{
			// add to get UTC
			utcSwitchTime = utcSwitchTime + (unsigned long)(tz->DaylightBias * 60L);
		} else
		{
			// subtract to get UTC
			utcSwitchTime = utcSwitchTime - (unsigned long)(tz->DaylightBias * (-60L));
		}
	}


	return utcSwitchTime;
}



/*
 .    Name:          TimeDateCompare()
 .
 .    Description:   Compares the time/date contents of two WCATIME
 .                   structures
 .       
 .    Returns:       -1 if time/date in 't1' < time/date in 't2'
 .                    0 if time/date in 't1' = time/date in 't2'
 .                    1 if time/date in 't1' > time/date in 't2'
 .
 .		Note:				does not consider daylight savings time in comparisons
 .
 */

int TimeDateCompare(WCATIME *t1, WCATIME *t2)
{
   if (t1->wYear > t2->wYear)
      return(1);
   else
   if (t1->wYear < t2->wYear)
      return(-1);
   else
   {
      if (t1->wMonth > t2->wMonth)
         return(1);
      else
      if (t1->wMonth < t2->wMonth)
         return(-1);
      else
      {
         if (t1->wDay > t2->wDay)
            return(1);
         else
         if (t1->wDay < t2->wDay)
            return(-1);
         else
            return(TimeCompare(t1,t2));
      }
   }
}



/*
 .    Name:          TimeCompare()
 .
 .    Description:   Compares the time contents of two SYSTEMTIME structures
 .       
 .    Returns:       -1 if time in 't1' < time in 't2'
 .                    0 if time in 't1' = time in 't2'
 .                    1 if time in 't1' > time in 't2'
 .
 .		Note:				does not consider daylight savings time in comparisons
 .    
 */
 
int TimeCompare(WCATIME *t1, WCATIME *t2)
{
   if (t1->wHour > t2->wHour)
      return(1);
   else
   if (t1->wHour < t2->wHour)
      return(-1);
   else
   {
      if (t1->wMinute > t2->wMinute)
         return(1);
      else
      if (t1->wMinute < t2->wMinute)
         return(-1);
      else
      {
         if (t1->wSecond > t2->wSecond)
            return(1);
         else
         if (t1->wSecond < t2->wSecond)
            return(-1);
         else
         {
            if (t1->wMilliseconds > t2->wMilliseconds)
               return(1);
            else
            if (t1->wMilliseconds < t2->wMilliseconds)
               return(-1);
            else
               return(0);
         }
      }
   }
}



/*
 .
 .    Name:          GetDaysInMonth()
 .
 .    Description:   Returns the number of days for the specified month
 .
 .    Note:          Handles leap year adjustment for February
 */

unsigned short GetDaysInMonth(unsigned short month, unsigned short year)
{
   if (month != 2)
      return(_daysMonthEachNonLeapYear[month]);
   else
      return(_daysMonthEachNonLeapYear[month] + IsLeapYear(year));
}


/*
 .    Name:          IsLeapYear()
 .
 .    Description:   Determines if the year is a leap year
 .
 .    Returns:       0 if not a leap year
 .                   1 if a leap year
 */

int IsLeapYear(unsigned short year)
{
   if ((year % 100) == 0)
   {
      /* the year is the beginning of a new century, must be evenly
         divisible by 400 to be a leap year */

      if ((year % 400) == 0)
         return(1);
      else
         return(0);
   }
   else
   {
      /* the year must be evenly divisible by four to be a leap year */

      if ((year % 4) == 0)
         return(1);
      else
         return(0);
   }
}



/*
 .    Name:          GetDayOfWeek()
 .
 .    Description:   Returns day of week (0 = Sunday, etc.)
 .
 .    In:            number of days elapsed since January 1, 1980
 .
 */

unsigned char GetDayOfWeek(unsigned long dayCnt)
{
   unsigned char dayTmp;

   dayTmp = (unsigned char)((dayCnt % 7L) + 1);
   if (dayTmp > 6) dayTmp -= 7;

   return(dayTmp);
}



/*
 .    Name:          AddTzToList()
 .
 .    Description:   Add a time zone record node to the end of the time
 .                   zone list.  Initializes the list if neccessary.
 .
 .    Returns:       TIME_ZONE_RECORD pointer to new node
 */

TIME_ZONE_LIST *AddTzToList(TIME_ZONE_LIST **tzList)
{
   TIME_ZONE_LIST *tzTmp;

   if (*tzList == NULL)
   {
      /* initialize head of list */

      tzTmp = (TIME_ZONE_LIST *)malloc(sizeof(TIME_ZONE_LIST));

      if (!tzTmp) return(NULL);

      tzTmp->prev = NULL;
      tzTmp->next = NULL;

      *tzList = tzTmp;

      return(tzTmp);
   }
   else
   {
      /* attach time zone record node to end of list */

      tzTmp = *tzList;

      while(tzTmp->next)
	  {
		  tzTmp = (TIME_ZONE_LIST *)tzTmp->next;
	  }

      tzTmp->next = malloc(sizeof(TIME_ZONE_LIST));

      if (!tzTmp->next) return(NULL);

      ((TIME_ZONE_LIST *)tzTmp->next)->prev = tzTmp;
      ((TIME_ZONE_LIST *)tzTmp->next)->next = NULL;

      return((TIME_ZONE_LIST *)tzTmp->next);
   }
}

/*
 .    Name:          FreeTzList()
 .
 .    Description:   Frees a linked TIME_ZONE_LIST
 .
 */

void FreeTzList(TIME_ZONE_LIST **tzList)
{
   TIME_ZONE_LIST *tzTrav;
   
   tzTrav = *tzList;

   while(tzTrav)
   {
      tzTrav = (TIME_ZONE_LIST *)tzTrav->next;

      free(*tzList);
   
      *tzList = tzTrav;
   }            
}




/*
 .    Name:          LoadTzFile()
 .
 .    Description:   Load time zones from TIME_ZONE_RECORD data file
 .
 */

unsigned short LoadTzFile(char *file, unsigned short *tzLoaded, TIME_ZONE_LIST **tzList)
{
   FILE              *inFile;
   errno_t            OpenError;

   TIME_ZONE_RECORD  tzRec;
   TIME_ZONE_LIST    *tzListNew;
   char              *x,*y;
   unsigned long     itemsReq;
   char              buffer[1024];

   *tzLoaded = 0;
   *tzList = NULL;

   OpenError = fopen_s(&inFile, file, "rt");
   if (OpenError)
	   return(FALSE);

   while(fgets(buffer,sizeof(buffer),inFile))
   {
      /* strip carriage return */

      buffer[strlen(buffer)-1] = '\0';

      x = strchr(buffer,'[');
      y = strchr(buffer,']');

      if (x && y) 
      {
         /* we have a new time zone definition */

         itemsReq = 0L;

         memset(buffer,0,sizeof(buffer));
      }

      x = strstr(buffer,"Name");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            strncpy_s(tzRec.tzName,sizeof(tzRec.tzName),y+1,sizeof(tzRec.tzName));
            
            itemsReq |= READ_NAME;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardBias");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardBias = atol(y+1);
            
            itemsReq |= READ_STDBIAS;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightBias");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightBias = atol(y+1);
            
            itemsReq |= READ_DYLBIAS;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"Bias");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.Bias = atol(y+1);
            
            itemsReq |= READ_BIAS;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.DayOfWeek");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wDayOfWeek = atoi(y+1);
            
            itemsReq |= READ_STDDAYOFWEEK;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Month");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wMonth = atoi(y+1);
            
            itemsReq |= READ_STDMONTH;

            memset(buffer,0,sizeof(buffer));
         }
      }
      
      x = strstr(buffer,"StandardDate.Day");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wDay = atoi(y+1);
            
            itemsReq |= READ_STDDAY;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Year");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wYear = atoi(y+1);
            
            itemsReq |= READ_STDYEAR;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Hour");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wHour = atoi(y+1);
            
            itemsReq |= READ_STDHOUR;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Minute");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wMinute = atoi(y+1);
            
            itemsReq |= READ_STDMINUTE;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Second");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wSecond = atoi(y+1);
            
            itemsReq |= READ_STDSECOND;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"StandardDate.Millisecond");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.StandardDate.wMilliseconds = atoi(y+1);
            
            itemsReq |= READ_STDMILLISECONDS;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.DayOfWeek");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wDayOfWeek = atoi(y+1);
            
            itemsReq |= READ_DYLDAYOFWEEK;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Month");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wMonth = atoi(y+1);
            
            itemsReq |= READ_DYLMONTH;

            memset(buffer,0,sizeof(buffer));
         }
      }
      
      x = strstr(buffer,"DaylightDate.Day");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wDay = atoi(y+1);
            
            itemsReq |= READ_DYLDAY;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Year");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wYear = atoi(y+1);
            
            itemsReq |= READ_DYLYEAR;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Hour");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wHour = atoi(y+1);
            
            itemsReq |= READ_DYLHOUR;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Minute");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wMinute = atoi(y+1);
            
            itemsReq |= READ_DYLMINUTE;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Second");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wSecond = atoi(y+1);
            
            itemsReq |= READ_DYLSECOND;

            memset(buffer,0,sizeof(buffer));
         }
      }

      x = strstr(buffer,"DaylightDate.Millisecond");

      if (x)
      {
         y = strchr(x,'=');

         if (y)
         {
            tzRec.tzInfo.DaylightDate.wMilliseconds = atoi(y+1);
            
            itemsReq |= READ_DYLMILLISECONDS;

            memset(buffer,0,sizeof(buffer));
         }
      }

      if (itemsReq == READ_ALL)
      {
         itemsReq |= READ_INVALIDATE;

         tzListNew = AddTzToList(tzList);

         if (!tzListNew) return(FALSE);

         memcpy(&tzListNew->tzRec,&tzRec,sizeof(TIME_ZONE_RECORD));

         (*tzLoaded)++;
      }
   }

   fclose(inFile);

   return(TRUE);
}




unsigned short ConvertWindowYear(unsigned char year)
{
	if (year < 80)
		return(2000 + year);
	else
		return(1900 + year);
}

unsigned short GetAmPmHour(unsigned short hr)
{
   if (hr > 12) return(hr - 12);
   if (hr == 0) 
      return(12);
   else
      return(hr);
}



void Win32ToWCATimeConvert(WCATIME *tdest, SYSTEMTIME *tsrc)
{
	tdest->wYear = tsrc->wYear;
	tdest->wMonth = tsrc->wMonth;
	tdest->wDayOfWeek = tsrc->wDayOfWeek;
	tdest->wDay = tsrc->wDay;
	tdest->wHour = tsrc->wHour;
	tdest->wMinute = tsrc->wMinute;
	tdest->wSecond = tsrc->wSecond;
	tdest->wMilliseconds = tsrc->wMilliseconds;

	/* assume no DST */

	tdest->wIsDstActive = FALSE;
}

void Win32ToWCATimeZoneInfo(WCA_TIME_ZONE_INFORMATION *tdest, TIME_ZONE_INFORMATION *tsrc)
{
	tdest->Bias         = tsrc->Bias;
	tdest->DaylightBias = tsrc->DaylightBias;
	tdest->StandardBias = tsrc->StandardBias;

	Win32ToWCATimeConvert(&tdest->DaylightDate,&tsrc->DaylightDate);
	Win32ToWCATimeConvert(&tdest->StandardDate,&tsrc->StandardDate);

	strcpy_s(tdest->DaylightName,sizeof(tdest->DaylightName),(char *)tsrc->DaylightName);
	strcpy_s(tdest->StandardName,sizeof(tdest->StandardName),(char *)tsrc->StandardName);
}


	
