
#include "stdafx.h"
#include <windef.h>
#include "PipeInterface.h"
#include "tcrcommon.h"
#include "DebugOutput.h"


//
//		Name:				::CPipeInterface()
//
//		Description:		Contructor initializes configuration class members.
//
//                          m_lpCmdLine gives the user the ability to pass in the name
//                          of the pipe server. This is normally not required.
//                          A null command line will find the default pipe server
//                          for the TCR1000. This will connect to a server local to the
//                          PC the monitor application is running on.
//							To connect to a remote pipe server the m_lpCmdLine will point
//							to a string with the name of the remote server.
//

CPipeInterface::CPipeInterface(LPTSTR m_lpCmdLine)
{
	DWORD			dwSize;
	char			*p;

	hPipeActive         = NULL;
	usePipeTransactions = TRUE;

	ZeroMemory(serverName,sizeof(serverName));
	dwSize = 0;
	p = m_lpCmdLine;
	while(*p && (*p != 32) && (dwSize < 1024))
	{
		*(serverName + dwSize) = *p;
		dwSize++;
		p++;
	}

	if (serverName[0])
		monitorSystemIsLocal = FALSE;
	else
	{
		strcpy_s(serverName,sizeof(serverName),"\\.");
		monitorSystemIsLocal = TRUE;
	}

}



//
//		Name:			PipeConnect()
//
//		Description:	Opens a connection to TCRSYNC's named pipe server on
//						the local or remote machine.  A named pipe server is
//						only present in the Windows NT/2000 implementation of
//						TCRSYNC time synchronization service.
//

BOOL CPipeInterface::PipeConnect(void)
{
	char *p;
	char tmp[1024];
	int iLength;
	char pipeName[1024];

	if (hPipeActive)
	{
		// already opened, exit
#ifdef _DEBUG_IPC
		DebugMsg("Pipe already connected, ignore PipeConnect");
#endif
		return TRUE;
	}

	strcpy_s(tmp,sizeof(tmp),serverName);

	// strip any prefixed backslashes

	p = tmp;

	iLength = 0;
	while(*p == '\\')
	{
		iLength++;
		p++;
	}

	// The max. size must be adjusted to not go beyond the end
	_strupr_s(p,sizeof(tmp)-iLength);

	strcpy_s(monitorSystem,sizeof(monitorSystem),p);

#ifdef _DEBUG_IPC
	DebugMsg("Monitor will watch computer %s",monitorSystem);
#endif

	sprintf_s(pipeName,sizeof(pipeName),"\\\\%s\\pipe\\%s",monitorSystem,TCRSYNC_IPC_NAME);

	// so that this program is compatible with Windows 95, we will not
	// utilize overlapped I/O operations

#ifdef _DEBUG_IPC
	DebugMsg("Open pipe: %s", pipeName);
#endif

	hPipeActive = CreateFile(	pipeName,
		GENERIC_READ|GENERIC_WRITE,
		FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		0,
		NULL);

	if (hPipeActive == INVALID_HANDLE_VALUE)
	{
#ifdef _DEBUG_ERROR
		DebugMsg("!!!unable to open named pipe instance %s",pipeName);
#endif

		hPipeActive = NULL;
		return(FALSE);
	}

	Sleep(500);

#ifdef _DEBUG_IPC
	DebugMsg("Pipe now opened");
#endif
	return(TRUE);
}


//
//		Name:			PipeTransaction()
//
//		Description:	Sends a transaction through the named pipe which is
//						connected to a TCRSYNC service running Win NT/2000.
//

BOOL CPipeInterface::PipeTransaction(DWORD cmd, PVOID txBuffer, DWORD txBuffSize,
	PVOID rxBuffer, DWORD rxBuffSize)
{
	BOOL            bConnect, bError;
	PIPE_CONTROL	pipeCmd;

	DWORD			cbWritten, cbRead, result, dwFilePointer;

	if (!hPipeActive)
	{
		bConnect = PipeConnect();

		if (!bConnect)
		{
#ifdef _DEBUG_ERROR
			DebugMsg("!!!Pipe Connect in pipe transaction failed");
#endif
			return FALSE;
		}

	}
#ifdef _DEBUG_IPC
	DebugMsg("Pipe trans cmd write: %d", (int)cmd);
#endif

	SetLastError(0);

	pipeCmd.cmd    = cmd;
	pipeCmd.length = txBuffSize;

	WriteFile(hPipeActive,(PVOID)&pipeCmd,sizeof(pipeCmd),&cbWritten,NULL);

	result = GetLastError();
	if ((result != ERROR_SUCCESS) || (cbWritten != sizeof(pipeCmd)))
	{
#ifdef _DEBUG_ERROR
		DebugMsg("!!!Pipe cmd write error: %ld wrote: %d desired: %d", (long)result, (int)cbWritten, (int)sizeof(pipeCmd));
#endif
		PipeShutdown();
		return(FALSE);
	}

	if (txBuffSize)
	{
		Sleep(200);

#ifdef _DEBUG_IPC
		DebugMsg("Pipe trans data write: %d len:%d", (int)cmd, (int)txBuffSize);
#endif

		WriteFile(hPipeActive,txBuffer,txBuffSize,&cbWritten,NULL);

		result = GetLastError();
		// for an unknown reason we get back an ERROR_INVALID_HANDLE even though the packet was sent.
		// Ignore this error code. The suspicion is that the service is using unicode
		// and this application does not. So, we are sending 8-bit bytes and the servioce is receiving
		// 16-bit unicode.
		if (((result != ERROR_INVALID_HANDLE) && (result != ERROR_SUCCESS)) || (!cbWritten))
		{
#ifdef _DEBUG_ERROR
			DebugMsg("!!!Pipe data write error: %ld written: %d", (long)result, (int)cbWritten);
#endif
			Sleep(5000);
			PipeShutdown();
			return(FALSE);
		}
	}

#ifdef _DEBUG_IPC
	DebugMsg("Pipe read: %d", cmd);
#endif

	ReadFile(hPipeActive,(PVOID)&pipeCmd,sizeof(pipeCmd),&cbRead,NULL);

	bError = FALSE;
	result = GetLastError();
	if ((result != ERROR_MORE_DATA) && (result != ERROR_SUCCESS) && (result != ERROR_INVALID_HANDLE))
		bError = TRUE;

	if (!cbRead)
		bError = TRUE;

	if (bError)
	{
#ifdef _DEBUG_ERROR
		DebugMsg("!!!Pipe read #1 failed error: %ld len:%d", (long)result, (int)cbRead);
#endif
		PipeShutdown();
		return(FALSE);
	}

	dwFilePointer = SetFilePointer(hPipeActive,cbRead,NULL,FILE_CURRENT);

	if (dwFilePointer == INVALID_SET_FILE_POINTER)
	{
#ifdef _DEBUG_ERROR
		DebugMsg("!!!Pipe SetFilePointer error: %ld len:%d", (long)dwFilePointer, (int)cbRead);
#endif
		PipeShutdown();
		return(FALSE);
	}


	if (pipeCmd.length)
	{
		if (!rxBuffer || !rxBuffSize)
		{
#ifdef _DEBUG_ERROR
			DebugMsg("!!!Pipe buffer empty: buffer pointer:%ld len:%d", (long)rxBuffer, (int)rxBuffSize);
#endif
			return(FALSE);
		}

#ifdef _DEBUG_IPC
		DebugMsg("Pipe read resp");
#endif

		ReadFile(hPipeActive,rxBuffer,pipeCmd.length,&cbRead,NULL);

		bError = FALSE;
		result = GetLastError();
		if ((result != ERROR_MORE_DATA) && (result != ERROR_SUCCESS) && (result != ERROR_INVALID_HANDLE))
			bError = TRUE;

		if ((!cbRead) || (cbRead != pipeCmd.length))
			bError = TRUE;

		if (bError)
		{
#ifdef _DEBUG_ERROR
			DebugMsg("!!!Pipe read #2 failed error: %ld received:%d requested:%d", (long)result, (int)cbRead, (int)pipeCmd.length);
#endif
			PipeShutdown();
			return(FALSE);
		}
	}

	return(TRUE);
}


//
//		Name:			PipeShutdown()
//
//		Description:	Closes a named-pipe handle (if open) to a TCRSYNC
//						service running on Win NT/2000.
//

void CPipeInterface::PipeShutdown(void)
{
#ifdef _DEBUG_IPC
	DebugMsg("Close Pipe");
#endif

	if (hPipeActive)
	{
		if (hPipeActive != INVALID_HANDLE_VALUE)
			CloseHandle(hPipeActive);
	}

	hPipeActive = NULL;
}



